#include "launchbutton.h"
#include <QPainter>
#include <QFontMetrics>
#include <QSizeF>
#include <KLocale>
#include <KConfigDialog>
#include <KColorScheme>
#include <KIO/NetAccess>
#include <KMessageBox>
#include <KFileDialog>
#include <QDomDocument>
#include <QDomElement>
#include <QLineEdit>
#include <QPlainTextEdit>

#include <plasma/widgets/pushbutton.h>
#include <plasma/theme.h>

LaunchButton::LaunchButton(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
    m_button(this),
    m_displayImage(""),
    m_execType(LAUNCHBUTTON_EXEC_TYPE_CUSTOM),
    m_customCommand(""),
    m_launcherCommand(""),
    m_dbusObject(""),
    m_dbusPath(""),
    m_dbusFunction(""),
    m_dbusArgs(""),
    m_useFileDialog(false),
    m_startingPath(""),
    m_fileFilter("*"),
    m_dbusConnection(QDBusConnection::sessionBus())
{
    // this will get us the standard applet background, for free!
    setBackgroundHints(NoBackground);
    setHasConfigurationInterface(true);
    m_button.setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    connect(&m_button, SIGNAL(clicked()), this, SLOT(execute()));
    resize(200, 200);
}


LaunchButton::~LaunchButton()
{
    if (hasFailedToLaunch()) {
        // Do some cleanup here
    } else {
        // Save settings
    }
}


void LaunchButton::constraintsEvent(Plasma::Constraints constraints)
{
    setBackgroundHints(NoBackground);
    if ( constraints & Plasma::SizeConstraint ) {
        m_button.setGeometry(rect());
    }
}

void LaunchButton::execute()
{
    QString arg = "";
    if (m_useFileDialog) {
        QStringList files = KFileDialog::getOpenFileNames(KUrl::fromPath(m_startingPath), m_fileFilter);
        if (files.isEmpty())
            return ;
        arg = "'" + files.replaceInStrings("'", "'\"'\"'").join("' '") + "'";
    }
    if (m_execType == LAUNCHBUTTON_EXEC_TYPE_CUSTOM) {
        QString cmd = m_customCommand + " " + arg;
        system(cmd.toAscii());
    } else if (m_execType == LAUNCHBUTTON_EXEC_TYPE_DBUS) {
        QString ifName = ui.dbusFunction->currentText();
        QString method;
        int pos = ui.dbusFunction->currentText().lastIndexOf(QLatin1Char('.'));
        if (pos == -1) {
            method = ifName;
            ifName.clear();
        } else {
            method = ifName.mid(pos + 1);
            ifName.truncate(pos);
        }
        QDBusInterface* iface = new QDBusInterface(ui.dbusObject->currentText(), ui.dbusPath->currentText(), ifName, m_dbusConnection);
        if (!iface->isValid()) {
            return ;
        }
        QVariantList params;
        if (!ui.dbusArgs->text().isEmpty()) {
            QStringList args = ui.dbusArgs->text().split(' ');
            const QMetaObject *mo = iface->metaObject();
            QByteArray match = method.toLatin1(); 
            match += '(';
            int i=mo->methodOffset();
            while (i < mo->methodCount() && !((QByteArray)mo->method(i).signature()).startsWith(match)) {
                i++;
            }
            if (i >= mo->methodCount()) {
                KMessageBox::error(0, i18n("Error: method %1 not found.", method));
                return ;
            }
            QMetaMethod mm = mo->method(i);
            
        }
        //Fill parameters
        QDBusMessage reply = iface->callWithArgumentList(QDBus::Block, method, params);
        if (reply.type() == QDBusMessage::ErrorMessage) {
            QDBusError err = reply;
            KMessageBox::error(0, i18n("Error %1: %2", err.name(), err.message()));
        } else if (reply.type() != QDBusMessage::ReplyMessage) {
            //KMessageBox::error(0, i18n("Reply: %1", reply.Message()));
        }
    } else {
        KMessageBox::error(0, i18n("This functionality isn't implemented yet."));
    }
}

void LaunchButton::dbusConnect()
{
    if (!m_dbusConnection.isConnected()) {
        KMessageBox::error(0, i18n("Could not connect to DBUS."));
        return ;
    }
}

void LaunchButton::dbusUpdateObjects()
{
    ui.dbusObject->clear();
    QDBusConnectionInterface* iface = m_dbusConnection.interface();
    if (!iface->isValid()) {
        return ;
    }
    const QStringList services = iface->registeredServiceNames();
    foreach (QString serviceName, services) {
        ui.dbusObject->addItem(serviceName);
    }
    
}

void LaunchButton::dbusAddPaths(QString path)
{
    QDBusInterface* iface = new QDBusInterface(ui.dbusObject->currentText(), path, QLatin1String("org.freedesktop.DBus.Introspectable"), m_dbusConnection);
    if (!iface->isValid()) {
        return ;
    }
    QDBusReply<QString> xml = iface->call(QLatin1String("Introspect"));
    if (!xml.isValid()) {
        return ;
    }
    ui.dbusPath->addItem(path);
    QDomDocument doc;
    doc.setContent(xml);
    QDomElement child = doc.documentElement().firstChildElement();
    while (!child.isNull()) {
        if (child.tagName() == QLatin1String("node")) {
            dbusAddPaths(path == "/" ? path + child.attribute(QLatin1String("name")) : path + QLatin1Char('/') + child.attribute(QLatin1String("name")));
        }
        child = child.nextSiblingElement();
    }
}

void LaunchButton::dbusUpdatePaths()
{
    ui.dbusPath->clear();
    dbusAddPaths("/");
}

void LaunchButton::dbusAddMethods(QString ifName)
{
    QDBusInterface* iface = new QDBusInterface(ui.dbusObject->currentText(), ui.dbusPath->currentText(), ifName, m_dbusConnection);
    if (!iface->isValid()) {
        return ;
    }
    const QMetaObject *mo = iface->metaObject(); 
    for (int i = mo->propertyOffset(); i < mo->propertyCount(); ++i) {
        QMetaProperty mp = mo->property(i);
        ui.dbusFunction->addItem(ifName + "." + mp.name());
    }
    for (int i = mo->methodOffset(); i < mo->methodCount(); ++i) {
        QMetaMethod mm = mo->method(i);
        QString signature = mm.signature(); 
        ui.dbusFunction->addItem(ifName + "." + signature);
    }
}

void LaunchButton::dbusUpdateFunctions()
{
    ui.dbusFunction->clear();
    QDBusInterface* iface = new QDBusInterface(ui.dbusObject->currentText(), ui.dbusPath->currentText(), QLatin1String("org.freedesktop.DBus.Introspectable"), m_dbusConnection);
    if (!iface->isValid()) {
        return ;
    }
    QDBusReply<QString> xml = iface->call(QLatin1String("Introspect"));
    if (!xml.isValid()) {
        return ;
    }
    QDomDocument doc;
    doc.setContent(xml);
    QDomElement child = doc.documentElement().firstChildElement();
    while (!child.isNull()) {
        if (child.tagName() == QLatin1String("interface")) {
            dbusAddMethods(child.attribute(QLatin1String("name")));
        }
        child = child.nextSiblingElement();
    }
}

void LaunchButton::dbusUpdatedObject()
{
    if (!ui.dbusObject->currentText().isEmpty()) {
        QDBusConnectionInterface* iface = m_dbusConnection.interface();
        if (!iface->isValid()) {
            return ;
        }
        if (!iface->isServiceRegistered(ui.dbusObject->currentText())) {
            KMessageBox::error(0, i18n("Service '%1' not registered.", ui.dbusObject->currentText()));
            return ;
        }
    }
    dbusUpdatePaths();
}

void LaunchButton::dbusUpdatedPath()
{
//    if (!ui.dbusPath->currentText().isEmpty() && !QDBusUtil::isValidObjectPath(ui.dbusPath->currentText())) {
//        KMessageBox::error(0, i18n("Invalid Path '%1'.", ui.dbusPath->currentText()));
//        return ;
//    }
    dbusUpdateFunctions();
}

void LaunchButton::createConfigurationInterface(KConfigDialog *parent)
{
    QWidget *widget = new QWidget();
    ui.setupUi(widget);
    connect(parent, SIGNAL(accepted()), this, SLOT(configAccepted())); 

    connect(ui.dbusObject, SIGNAL(currentIndexChanged(int)), this, SLOT(dbusUpdatedObject())); 
    connect(ui.dbusPath, SIGNAL(currentIndexChanged(int)), this, SLOT(dbusUpdatedPath())); 

    ui.displayImage->setUrl(KUrl::fromPath(m_displayImage));

    ui.useCustomCommand->setChecked(m_execType == LAUNCHBUTTON_EXEC_TYPE_CUSTOM);
    ui.customCommand->setText(m_customCommand);

    ui.useLauncher->setChecked(m_execType == LAUNCHBUTTON_EXEC_TYPE_LAUNCHER);
    ui.launcherCommand->setText(m_launcherCommand);

    ui.useDBUS->setChecked(m_execType == LAUNCHBUTTON_EXEC_TYPE_DBUS);
    ui.dbusFunction->clear();
    ui.dbusPath->clear();
    ui.dbusObject->clear();
    //fill objects from DBUS
    dbusUpdateObjects();
    //find object then set index
    int i = ui.dbusObject->findText(m_dbusObject);
    ui.dbusObject->setCurrentIndex(i);
    if (m_execType == LAUNCHBUTTON_EXEC_TYPE_DBUS && !m_dbusObject.isEmpty() &&  i < 0) {
        KMessageBox::error(0, i18n("The DBUS object '%1' isn't found, please cancel the dialog and load it first.", m_dbusObject));
    }
    if (i >= 0) {
        //fill paths from DBUS
        dbusUpdatePaths();
        i = ui.dbusPath->findText(m_dbusPath);
        ui.dbusPath->setCurrentIndex(i);
        if (i >= 0) {
            //fill functions from DBUS
            dbusUpdateFunctions();
            i = ui.dbusFunction->findText(m_dbusFunction);
            ui.dbusFunction->setCurrentIndex(i);
            if (i >= 0) {
                ui.dbusArgs->setText(m_dbusArgs);
            } else {
                ui.dbusArgs->setText("");
            }
        } else {
            ui.dbusFunction->setCurrentIndex(i);
            ui.dbusArgs->setText("");
        }
    } else {
        ui.dbusPath->setCurrentIndex(i);
        ui.dbusFunction->setCurrentIndex(i);
        ui.dbusArgs->setText("");
    }

    ui.useFileDialog->setChecked(m_useFileDialog);
    ui.startingPath->setMode(KFile::Directory);
    ui.startingPath->setUrl(KUrl::fromPath(m_startingPath));
    ui.fileFilter->setPlainText(m_fileFilter);

    parent->addPage(widget, i18n("General"), icon());
}


void LaunchButton::configAccepted()
{
    KConfigGroup cg = config();

    if (ui.useCustomCommand->isChecked())
        m_execType = LAUNCHBUTTON_EXEC_TYPE_CUSTOM;
    else if (ui.useLauncher->isChecked())
        m_execType = LAUNCHBUTTON_EXEC_TYPE_LAUNCHER;
    else if (ui.useDBUS->isChecked())
        m_execType = LAUNCHBUTTON_EXEC_TYPE_DBUS;
    else
        m_execType = 0;
    cg.writeEntry("execType", m_execType);
    
    m_customCommand = ui.customCommand->text();
    cg.writeEntry("customCommand", m_customCommand);
    
    m_launcherCommand = ui.launcherCommand->text();
    cg.writeEntry("launcherCommand", m_launcherCommand);
    
    m_dbusObject = ui.dbusObject->currentText();
    cg.writeEntry("dbusObject", m_dbusObject);
    
    m_dbusPath = ui.dbusPath->currentText();
    cg.writeEntry("dbusPath", m_dbusPath);
    
    m_dbusFunction = ui.dbusFunction->currentText();
    cg.writeEntry("dbusFunction", m_dbusFunction);
    
    m_dbusArgs = ui.dbusArgs->text();
    cg.writeEntry("dbusArgs", m_dbusArgs);
    
    if (ui.displayImage->url().isEmpty() || KIO::NetAccess::exists(ui.displayImage->url(), KIO::NetAccess::SourceSide, NULL)) {
        m_displayImage = ui.displayImage->url().path();
        m_button.setImage(m_displayImage);
        cg.writeEntry ("displayImage", m_displayImage);
    } else {
        KMessageBox::error(0, i18n("The given image could not be loaded. The image will not be changed."));
    }
    
    m_useFileDialog = ui.useFileDialog->isChecked();
    cg.writeEntry("useFileDialog", m_useFileDialog);
    
    m_startingPath = ui.startingPath->url().path();
    cg.writeEntry("startingPath", m_startingPath);
    
    m_fileFilter = ui.fileFilter->toPlainText();
    cg.writeEntry("fileFilter", m_fileFilter);
}


void LaunchButton::init()
{
    KConfigGroup cg = config();
    
    m_displayImage = cg.readEntry("displayImage", "");

    m_execType = cg.readEntry("execType", LAUNCHBUTTON_EXEC_TYPE_CUSTOM);
    m_customCommand = cg.readEntry("customCommand", "");
    m_launcherCommand = cg.readEntry("launcherCommand", "");
    m_dbusObject = cg.readEntry("dbusObject", "");
    m_dbusPath = cg.readEntry("dbusPath", "");
    m_dbusFunction = cg.readEntry("dbusFunction", "");
    m_dbusArgs = cg.readEntry("dbusArgs", "");
    
    m_useFileDialog = cg.readEntry("useFileDialog", false);
    m_startingPath = cg.readEntry("startingPath", "");
    m_fileFilter = cg.readEntry("fileFilter", "*");
    
    if (m_displayImage.isEmpty() || KIO::NetAccess::exists(m_displayImage, KIO::NetAccess::SourceSide, NULL)) {
        m_button.setImage(m_displayImage);
    } else {
        KMessageBox::error(0, i18n("The image could not be found."));
    }
}


void LaunchButton::paintInterface(QPainter *p,
        const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    p->setRenderHint(QPainter::SmoothPixmapTransform);
    p->setRenderHint(QPainter::Antialiasing);
}

#include "launchbutton.moc"
